import pytest
import os
from click.testing import CliRunner
from threading import Thread
from .api_mocks import upload_logs, upsert_run
from freezegun import freeze_time

import wandb
import time


def mock_stop(*args):
    pass


def test_watches_for_all_changes(mocker):
    with CliRunner().isolated_filesystem():
        api = mocker.MagicMock()
        sync = wandb.Sync(api, "test", dir='.')
        sync.stop = mock_stop
        sync.watch(['*'])
        with open("some_file.h5", "w") as f:
            f.write("My great changes")
        # Fuck if I know why this makes shit work...
        time.sleep(1)
        assert api.upsert_run.called
        assert api.push.called


def test_watches_for_specific_change(mocker):
    with CliRunner().isolated_filesystem():
        api = mocker.MagicMock()
        sync = wandb.Sync(api, "test", dir='.')
        sync.stop = mock_stop
        sync.watch(["rad.txt"])
        with open("rad.txt", "a") as f:
            f.write("something great")
        time.sleep(1)
        assert api.push.called


def test_watches_for_subdir_change(mocker):
    with CliRunner().isolated_filesystem():
        api = mocker.MagicMock()
        sync = wandb.Sync(api, "test", dir='.')
        sync.stop = mock_stop
        sync.watch(["./subdir/*.txt"])
        os.mkdir('subdir')
        with open("subdir/rad.txt", "a") as f:
            f.write("something great")
        time.sleep(1)
        assert api.push.called


def test_ignores_hidden_folders(mocker):
    with CliRunner().isolated_filesystem():
        api = mocker.MagicMock()
        sync = wandb.Sync(api, "test", dir='.')
        sync.stop = mock_stop
        sync.watch(["*"])
        os.mkdir('.subdir')
        with open(".subdir/rad.txt", "a") as f:
            f.write("something great")
        time.sleep(1)
        assert not api.push.called


def test_watches_for_glob_change(mocker):
    with CliRunner().isolated_filesystem():
        api = mocker.MagicMock()
        sync = wandb.Sync(api, "test", dir='.')
        sync.stop = mock_stop
        sync.watch(["*.txt"])
        with open("file.txt", "a") as f:
            f.write("great")
        time.sleep(1)
        assert api.push.called

# def test_syncs_log(mocker, upload_logs, upsert_run, request_mocker):
#    with CliRunner().isolated_filesystem():
#        api = wandb.Api()
#        run_mock = upsert_run(request_mocker)
#        with freeze_time("1981-12-09 12:00:01"):
#            sync = wandb.Sync(api, dir='.')
#        log_mock = upload_logs(request_mocker, sync.run_id)
#        sync.stop = mock_stop
#        sync.watch('*')
#        assert run_mock.called
#        print("My logger")
#        print("1")
#        print("2")
#        print("3")
#        print("4th and final")
#        time.sleep(1)
#        assert log_mock.called



import psutil
import os
import stat
import sys
import time
import traceback
from tempfile import NamedTemporaryFile
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
from shortuuid import ShortUUID
import atexit
from .config import Config
import logging
import threading
from six.moves import queue
import socket
import click
from wandb import __stage_dir__, Error
from wandb import streaming_log
from wandb import util
from .api import BinaryFilePolicy, CRDedupeFilePolicy
from .wandb_run import Run
logger = logging.getLogger(__name__)


def editor(content='', marker='# Before we start this run, enter a brief description. (to skip, direct stdin to dev/null: `python train.py < /dev/null`)\n'):
    message = click.edit(content + '\n\n' + marker)
    if message is None:
        return None
    return message.split(marker, 1)[0].rstrip('\n')


class OutStreamTee(object):
    """Tees a writable filelike object.

    writes/flushes to the passed in stream will go to the stream
    and a second stream.
    """

    def __init__(self, stream, second_stream):
        """Constructor.

        Args:
            stream: stream to tee.
            second_stream: stream to duplicate writes to.
        """
        self._orig_stream = stream
        self._second_stream = second_stream
        self._queue = queue.Queue()
        self._thread = threading.Thread(target=self._thread_body)
        self._thread.daemon = True
        self._thread.start()

    def _thread_body(self):
        while True:
            item = self._queue.get()
            if item is None:
                break
            self._second_stream.write(item)

    def write(self, message):
        #print('writing to orig: ', self._orig_stream)
        self._orig_stream.write(message)
        # print('queueing')
        self._queue.put(message)

    def flush(self):
        self._orig_stream.flush()

    def close(self):
        self._queue.put(None)


class ExitHooks(object):
    def __init__(self):
        self.exit_code = None
        self.exception = None

    def hook(self):
        self._orig_exit = sys.exit
        self._orig_excepthook = sys.excepthook
        sys.exit = self.exit
        sys.excepthook = self.excepthook

    def exit(self, code=0):
        self.exit_code = code
        self._orig_exit(code)

    def excepthook(self, exc_type, exc, *args):
        self.exception = exc
        self._orig_excepthook(exc_type, exc, *args)


class FileTailer(object):
    def __init__(self, path, on_read_fn, binary=False):
        self._path = path
        mode = 'r'
        if binary:
            mode = 'rb'
        self._file = open(path, mode)
        self._on_read_fn = on_read_fn
        self._thread = threading.Thread(target=self._thread_body)
        self._thread.start()

    def _thread_body(self):
        while True:
            data = self._file.read(1024)
            if data == '':
                time.sleep(1)
            else:
                self._on_read_fn(data)


class FileEventHandler(object):
    def __init__(self, file_path, save_name, api):
        self.file_path = file_path
        self.save_name = save_name
        self._api = api

    def on_created(self):
        pass

    def on_modified(self):
        pass


class FileEventHandlerOverwrite(FileEventHandler):
    def __init__(self, file_path, save_name, api, project, run_id, *args, **kwargs):
        super(FileEventHandlerOverwrite, self).__init__(
            file_path, save_name, api, *args, **kwargs)
        self._project = project
        self._run_id = run_id
        self._tailer = None

    def on_created(self):
        self.on_modified()

    def on_modified(self):
        print("Pushing %s" % self.file_path)
        with open(self.file_path, 'rb') as f:
            self._api.push(self._project, {self.save_name: f}, run=self._run_id,
                           progress=False)


class FileEventHandlerTextStream(FileEventHandler):
    def __init__(self, *args, **kwargs):
        super(FileEventHandlerTextStream, self).__init__(*args, **kwargs)
        self._tailer = None

    def on_created(self):
        if self._tailer:
            logger.error(
                'Streaming file created twice in same run: %s', self.file_path)
            return
        self._setup()

    def on_modified(self):
        if self._tailer:
            return
        self._setup()

    def _setup(self):
        fsapi = self._api.get_file_stream_api()
        pusher = streaming_log.TextStreamPusher(fsapi, self.save_name)

        def on_read(data):
            pusher.write(data)

        self._tailer = FileTailer(self.file_path, on_read)


class FileEventHandlerBinaryStream(FileEventHandler):
    def __init__(self, *args, **kwargs):
        super(FileEventHandlerBinaryStream, self).__init__(*args, **kwargs)
        self._tailer = None

    def on_created(self):
        if self._tailer:
            logger.error(
                'Streaming file created twice in same run: %s', self.file_path)
            return
        self._setup()

    def on_modified(self):
        if self._tailer:
            return
        self._setup()

    def _setup(self):
        fsapi = self._api.get_file_stream_api()

        def on_read(data):
            fsapi.push(self.save_name, data)

        self._tailer = FileTailer(self.file_path, on_read, binary=True)


class Sync(object):
    """Watches for files to change and automatically pushes them
    """

    def __init__(self, api, run_id, config=None, project=None, tags=[], datasets=[], description=None, dir=None):
        # 1.6 million 6 character combinations
        self._run_id = run_id
        self._project = project or api.settings("project")
        self._entity = api.settings("entity")
        logger.debug("Initialized sync for %s/%s", self._project, self._run_id)
        self._dpath = os.path.join(__stage_dir__, 'description.md')
        self._description = description or (os.path.exists(self._dpath) and open(
            self._dpath).read()) or os.getenv('WANDB_DESCRIPTION')
        try:
            self.tty = sys.stdin.isatty() and os.getpgrp() == os.tcgetpgrp(sys.stdout.fileno())
        except OSError:
            self.tty = False
        if not os.getenv('DEBUG') and not self._description and self.tty:
            self._description = editor()
            if self._description is None:
                sys.stderr.write('No description provided, aborting run.\n')
                sys.exit(1)
        self._proc = psutil.Process(os.getpid())
        self._api = api
        self._tags = tags
        self._handler = PatternMatchingEventHandler()
        self._handler.on_created = self.on_file_created
        self._handler.on_modified = self.on_file_modified
        base_url = api.settings('base_url')
        if base_url.endswith('.dev'):
            base_url = 'http://app.dev'
        self.url = "{base}/{entity}/{project}/runs/{run}".format(
            project=self._project,
            entity=self._entity,
            run=self._run_id,
            base=base_url
        )
        self._hooks = ExitHooks()
        self._hooks.hook()
        self._observer = Observer()
        if dir is None:
            self._watch_dir = os.path.join(
                __stage_dir__, 'run-%s' % self._run_id)
            util.mkdir_exists_ok(self._watch_dir)
        else:
            self._watch_dir = os.path.abspath(dir)

        self._observer.schedule(self._handler, self._watch_dir, recursive=True)

        if config is None:
            config = Config()
        self._config = config

        self._event_handlers = {}

    def watch(self, files):
        try:
            # TODO: better failure handling
            self._api.upsert_run(name=self._run_id, project=self._project, entity=self._entity,
                                 config=self._config.__dict__, description=self._description, host=socket.gethostname())
            self._handler._patterns = [
                os.path.join(self._watch_dir, os.path.normpath(f)) for f in files]
            # Ignore hidden files/folders
            self._handler._ignore_patterns = ['*/.*']
            if os.path.exists(__stage_dir__ + "diff.patch"):
                self._api.push("{project}/{run}".format(
                    project=self._project,
                    run=self._run_id
                ), {"diff.patch": open(__stage_dir__ + "diff.patch", "rb")})
            self._observer.start()

            print("Syncing %s" % self.url)

            self._api.get_file_stream_api().set_file_policy(
                'output.log', CRDedupeFilePolicy())
            # Tee stdout/stderr into our TextOutputStream, which will push lines to the cloud.
            self._stdout_stream = streaming_log.TextStreamPusher(
                self._api.get_file_stream_api(), 'output.log', prepend_timestamp=True)
            sys.stdout = OutStreamTee(sys.stdout, self._stdout_stream)
            self._stderr_stream = streaming_log.TextStreamPusher(
                self._api.get_file_stream_api(), 'output.log', line_prepend='ERROR',
                prepend_timestamp=True)
            sys.stderr = OutStreamTee(sys.stderr, self._stderr_stream)

            self._stdout_stream.write(" ".join(psutil.Process(
                os.getpid()).cmdline()) + "\n\n")

            logger.debug("Swapped stdout/stderr")

            atexit.register(self.stop)
        except KeyboardInterrupt:
            self.stop()
        except Error:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            print("!!! Fatal W&B Error: %s" % exc_value)
            lines = traceback.format_exception(
                exc_type, exc_value, exc_traceback)
            logger.error('\n'.join(lines))

    def stop(self):
        # This is a a heuristic delay to catch files that were written just before
        # the end of the script. This is unverified, but theoretically the file
        # change notification process used by watchdog (maybe inotify?) is
        # asynchronous. It's possible we could miss files if 10s isn't long enough.
        # TODO: Guarantee that all files will be saved.
        print("Script ended, waiting for final file modifications.")
        time.sleep(10.0)
        # self.log.tempfile.flush()
        print("Pushing log")
        slug = "{project}/{run}".format(
            project=self._project,
            run=self._run_id
        )
        # self._api.push(
        #    slug, {"training.log": open(self.log.tempfile.name, "rb")})
        os.path.exists(self._dpath) and os.remove(self._dpath)
        print("Synced %s" % self.url)
        self._stdout_stream.close()
        self._stderr_stream.close()
        self._api.get_file_stream_api().finish(self._hooks.exception)
        try:
            self._observer.stop()
            self._observer.join()
        # TODO: py2 TypeError: PyCObject_AsVoidPtr called with null pointer
        except TypeError:
            pass
        # TODO: py3 SystemError: <built-in function stop> returned a result with an error set
        except SystemError:
            pass

    def _get_handler(self, file_path, save_name):
        if save_name not in self._event_handlers:
            if save_name == 'wandb-history.jsonl':
                self._event_handlers['wandb-history.jsonl'] = FileEventHandlerTextStream(
                    file_path, 'wandb-history.jsonl', self._api)
            elif 'tfevents' in save_name:
                # TODO: This is hard-coded, but we want to give users control
                # over streaming files (or detect them).
                self._api.get_file_stream_api().set_file_policy(save_name,
                                                                BinaryFilePolicy())
                self._event_handlers[save_name] = FileEventHandlerBinaryStream(
                    file_path, save_name, self._api)
            else:
                self._event_handlers[save_name] = FileEventHandlerOverwrite(
                    file_path, save_name, self._api, self._project, self._run_id)
        return self._event_handlers[save_name]

    # TODO: limit / throttle the number of adds / pushes
    def on_file_created(self, event):
        if os.stat(event.src_path).st_size == 0 or os.path.isdir(event.src_path):
            return None
        save_name = os.path.relpath(event.src_path, self._watch_dir)
        self._get_handler(event.src_path, save_name).on_created()

    # TODO: is this blocking the main thread?
    def on_file_modified(self, event):
        if os.stat(event.src_path).st_size == 0 or os.path.isdir(event.src_path):
            return None
        save_name = os.path.relpath(event.src_path, self._watch_dir)
        self._get_handler(event.src_path, save_name).on_modified()

    @property
    def source_proc(self):
        mode = os.fstat(0).st_mode
        if not stat.S_ISFIFO(mode):
            # stdin is not a pipe
            return None
        else:
            source = self._proc.parent().children()[0]
            return None if source == self._proc else source

    def echo(self):
        print(sys.stdin.read())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_results
----------------------------------

Tests for the `wandb.Results` module.
"""
import pytest
from wandb import Results
from .api_mocks import *
from click.testing import CliRunner

def test_push_success(request_mocker, upload_url, query_project):
    #mock = query_project(request_mocker)
    payload = project("test", files={
        'edges': [
            {'node': {
                'name': 'results.csv',
                'url': 'https://results.csv',
                'md5': 'fakemd5'
            }}]})
    success_or_failure({'model': payload})(request_mocker)
    #upload_url(request_mocker)
    mock = request_mocker.register_uri('PUT', 'https://results.csv', status_code=200)
    with CliRunner().isolated_filesystem():
        with Results("test/test") as res:
            res.write(input="Test", output=False, truth=True, loss=0.89)
    assert mock.called


from git import Repo, exc
import os

class GitRepo(object):
    def __init__(self, root=None, remote="origin", lazy=True):
        self.remote_name = remote
        self.root = root
        self._repo = None
        if not lazy:
            self.repo

    @property
    def repo(self):
        if self._repo is None:
            if self.remote_name is None:
                self._repo = False
            else:
                try:
                    self._repo = Repo(self.root or os.getcwd(), search_parent_directories=True)
                except exc.InvalidGitRepositoryError:
                    self._repo = False
        return self._repo
    
    @property
    def enabled(self):
        return self.repo

    @property
    def dirty(self):
        if not self.repo:
            return False
        return self.repo.is_dirty()

    @property
    def last_commit(self):
        if not self.repo:
            return None
        return self.repo.head.commit.hexsha

    @property
    def branch(self):
        if not self.repo:
            return None
        return self.repo.head.ref.name

    @property
    def remote(self):
        if not self.repo:
            return None
        try:
            return self.repo.remotes[self.remote_name]
        except IndexError:
            return None

    @property
    def remote_url(self):
        if not self.remote:
            return None
        return self.remote.url

    def tag(self, name, message):
        try:
            return self.repo.create_tag("wandb/"+name, message=message, force=True)
        except exc.GitCommandError:
            print("Failed to tag repository.")
            return None

    def push(self, name):
        if self.remote:
            try:
                return self.remote.push("wandb/"+name, force=True)
            except exc.GitCommandError:
                return None



import yaml
import os
import sys
import logging
from .api import Error
from wandb import __stage_dir__


def boolify(s):
    if s.lower() == 'none':
        return None
    if s.lower() == 'true':
        return True
    if s.lower() == 'false':
        return False
    raise ValueError("Not a boolean")


class Config(dict):
    """Creates a W&B config object.

    The object is an enhanced `dict`.  You can access keys via instance methods or
    as you would a regular `dict`.  The object first looks for a `config-defaults.yaml` file
    in the current directory.  It then looks for environment variables pre-pended
    with "WANDB_".  Lastly it overrides any key found in command line arguments.

    Using the config objects enables W&B to track all configuration parameters used
    in your training runs.

    Args:
        config(:obj:`dict`, optional): Key value pairs from your existing code that
        you would like to track.  You can also pass in objects that respond to `__dict__`
        such as from argparse.
    """

    def __init__(self, config={}):
        if not isinstance(config, dict):
            try:
                # for tensorflow flags
                if "__flags" in dir(config):
                    config._parse_flags()
                    config = config.__dict__['__flags']
                else:
                    config = vars(config)
            except TypeError:
                raise TypeError(
                    "config must be a dict or have a __dict__ attribute.")
        dict.__init__(self, {})
        self._descriptions = {}
        # we only persist when _external is True
        self._external = False
        self.load_defaults()
        self.load_env()
        self.load_overrides()
        for key in config:
            self[key] = config[key]
        self._external = True
        self.persist(overrides=True)

    @property
    def config_dir(self):
        """The config directory holding the latest configuration"""
        return os.path.join(os.getcwd(), __stage_dir__)

    @property
    def defaults_path(self):
        """Where to find the default configuration"""
        return os.getcwd() + "/config-defaults.yaml"

    @property
    def keys(self):
        """All keys in the current configuration"""
        return [key for key in self if not key.startswith("_")]

    def desc(self, key):
        """The description of a given key"""
        return self._descriptions.get(key)

    def convert(self, ob):
        """Type casting for Boolean, None, Int and Float"""
        # TODO: deeper type casting
        if isinstance(ob, dict) or isinstance(ob, list):
            return ob
        for fn in (boolify, int, float):
            try:
                return fn(str(ob))
            except ValueError:
                pass
        return str(ob)

    def load_json(self, json):
        """Loads existing config from JSON"""
        for key in json:
            self[key] = json[key].get('value')
            self._descriptions[key] = json[key].get('desc')

    def load_defaults(self):
        """Load defaults from YAML"""
        if os.path.exists(self.defaults_path):
            try:
                defaults = yaml.load(open(self.defaults_path))
            except yaml.parser.ParserError:
                raise Error("Invalid YAML in config-defaults.yaml")
            if defaults:
                for key in defaults:
                    if key == "wandb_version":
                        continue
                    self[key] = defaults[key].get('value')
                    self._descriptions[key] = defaults[key].get('desc')
        else:
            logging.info(
                "Couldn't load default config, run `wandb config init` in this directory")

    def load_overrides(self):
        """Load overrides from command line arguments"""
        for arg in sys.argv:
            key_value = arg.split("=")
            if len(key_value) == 2:
                key = key_value[0].replace("--", "")
                if self.get(key):
                    self[key] = self.convert(key_value[1])

    def load_env(self):
        """Load overrides from the environment"""
        for key in [key for key in os.environ if key.startswith("WANDB_CONFIG_")]:
            value = os.environ[key]
            self[key.replace("WANDB_CONFIG_", "").lower()
                 ] = self.convert(value)

    def persist(self, overrides=False):
        """Stores the current configuration for pushing to W&B"""
        if overrides:
            path = "{dir}/latest.yaml".format(dir=self.config_dir)
        else:
            path = self.defaults_path
        try:
            with open(path, "w") as defaults_file:
                defaults_file.write(str(self))
            return True
        except IOError:
            logging.warn(
                "Unable to persist config, no wandb directory exists.  Run `wandb config init` in this directory.")
            return False

    def __getitem__(self, name):
        return super(Config, self).__getitem__(name)

    def __setitem__(self, key, value):
        # TODO: this feels gross
        if key.endswith("_desc"):
            parts = key.split("_")
            parts.pop()
            self._descriptions["_".join(parts)] = str(value)
        else:
            # TODO: maybe don't convert, but otherwise python3 dumps unicode
            super(Config, self).__setitem__(key, self.convert(value))
            if not key.startswith("_") and self._external:
                self.persist(overrides=True)
            return value

    def __getattr__(self, name):
        return self.get(name)

    def update(self, params):
        if not isinstance(params, dict):
            try:
                # for tensorflow flags
                if "__flags" in dir(params):
                    if not params.__parsed:
                        params._parse_flags()
                    params = params.__flags
                else:
                    params = vars(params)
            except TypeError:
                raise TypeError(
                    "config must be a dict or have a __dict__ attribute.")
        super(Config, self).update(params)

    __setattr__ = __setitem__

    @property
    def __dict__(self):
        defaults = {}
        for key in self.keys:
            defaults[key] = {'value': self[key],
                             'desc': self._descriptions.get(key)}
        return defaults

    def __repr__(self):
        rep = "\n".join([str({
            'key': key,
            'desc': self._descriptions.get(key),
            'value': self[key]
        }) for key in self.keys])
        return rep

    def __str__(self):
        s = "wandb_version: 1\n\n"
        if self.__dict__:
            s += yaml.dump(self.__dict__, default_flow_style=False)
        return s



from gql import Client, gql
from gql.client import RetryError
from gql.transport.requests import RequestsHTTPTransport
import os
import requests
import ast
from six.moves import configparser
from functools import wraps
import logging
import hashlib
import os
import json
import yaml
from wandb import __version__, __stage_dir__, GitRepo
from wandb import util
import base64
import binascii
import click
import collections
import itertools
import logging
import requests
from six.moves import queue
import threading
import time

logger = logging.getLogger(__name__)


def IDENTITY(monitor):
    """A default callback for the Progress helper"""
    return monitor


class Progress(object):
    """A helper class for displaying progress"""

    def __init__(self, file, callback=None):
        self.file = file
        self.callback = callback or IDENTITY
        self.bytes_read = 0
        self.len = os.fstat(file.fileno()).st_size

    def read(self, size=-1):
        """Read bytes and call the callback"""
        bites = self.file.read(size)
        self.bytes_read += len(bites)
        self.callback(len(bites))
        return bites


class Error(Exception):
    """Base W&B Error"""
    # For python 2 support

    def encode(self, encoding):
        return self.message


class CommError(Error):
    """Error communicating with W&B"""
    pass


class UsageError(Error):
    """API Usage Error"""
    pass


def normalize_exceptions(func):
    """Function decorator for catching common errors and re-raising as wandb.Error"""
    @wraps(func)
    def wrapper(*args, **kwargs):
        message = "Whoa, you found a bug."
        try:
            return func(*args, **kwargs)
        except requests.HTTPError as err:
            raise CommError(err.response)
        except RetryError as err:
            if "response" in dir(err.last_exception) and err.last_exception.response is not None:
                message = err.last_exception.response.json().get(
                    'errors', [{'message': message}])[0]['message']
            else:
                message = err.last_exception
            raise CommError(message)
        except Exception as err:
            # gql raises server errors with dict's as strings...
            payload = err.args[0]
            if str(payload).startswith("{"):
                message = ast.literal_eval(str(payload))["message"]
            else:
                message = str(err)
            if os.getenv("DEBUG") == "true":
                raise err
            else:
                raise CommError(message)
    return wrapper


class Api(object):
    """W&B Api wrapper

    Note:
        Settings are automatically overridden by looking for
        a `wandb/settings` file in the current working directory or it's parent
        directory.  If none can be found, we look in the current users home
        directory.

    Args:
        default_settings(:obj:`dict`, optional): If you aren't using a settings
        file or you wish to override the section to use in the settings file
        Override the settings here.
    """

    def __init__(self, default_settings=None, load_settings=True):
        self.default_settings = {
            'section': "default",
            'entity': "models",
            'run': "latest",
            'git_remote': "origin",
            'git_tag': False,
            'base_url': "https://api.wandb.ai"
        }
        self.default_settings.update(default_settings or {})
        self._settings = None
        self.retries = 3
        self._settings_parser = configparser.ConfigParser()
        self.tagged = False
        if load_settings:
            potential_settings_paths = [
                os.path.expanduser('~/.wandb/settings')
            ]
            if __stage_dir__ is not None:
                potential_settings_paths.append(
                    os.path.join(os.getcwd(), __stage_dir__, 'settings'))
            files=self._settings_parser.read(potential_settings_paths)
            self.settings_file=files[0] if len(files) > 0 else "Not found"
        else:
            self.settings_file="Not found"
        self.git=GitRepo(remote = self.settings("git_remote"))
        self._commit=self.git.last_commit
        if self.git.dirty:
            self.git.repo.git.execute(['git', 'diff'], output_stream = open(
                __stage_dir__ + 'diff.patch', 'wb'))
        self.client=Client(
            retries = self.retries,
            transport = RequestsHTTPTransport(
                headers={'User-Agent': self.user_agent},
                use_json=True,
                auth=("api", self.api_key),
                url='%s/graphql' % self.settings('base_url')
            )
        )
        self._current_run_id=None
        self._file_stream_api=None

    def set_current_run_id(self, run_id):
        self._current_run_id=run_id

    @property
    def current_run_id(self):
        return self._current_run_id

    @property
    def user_agent(self):
        return 'W&B Client %s' % __version__

    @property
    def api_key(self):
        auth=requests.utils.get_netrc_auth(self.settings()['base_url'])
        if auth:
            key=auth[-1]
        else:
            key=os.environ.get("WANDB_API_KEY")
        return key

    def settings(self, key = None, section = None):
        """The settings overridden from the wandb/settings file.

        Args:
            key (str, optional): If provided only this setting is returned
            section (str, optional): If provided this section of the setting file is
            used, defaults to "default"

        Returns:
            A dict with the current settings

                {
                    "entity": "models",
                    "base_url": "https://api.wandb.ai",
                    "project": None
                }
        """
        if not self._settings:
            self._settings=self.default_settings.copy()
            section=section or self._settings['section']
            try:
                if section in self._settings_parser.sections():
                    for option in self._settings_parser.options(section):
                        self._settings[option]=self._settings_parser.get(
                            section, option)
            except configparser.InterpolationSyntaxError:
                print("WARNING: Unable to parse settings file")
            self._settings["project"]=self._settings.get(
                "project", os.environ.get("WANDB_PROJECT"))
            self._settings["entity"] = self._settings.get(
                "entity", os.environ.get("WANDB_ENTITY"))
            self._settings["base_url"] = self._settings.get(
                "base_url", os.environ.get("WANDB_BASE_URL"))
        return self._settings if key is None else self._settings[key]

    def parse_slug(self, slug, project=None, run=None):
        if slug and "/" in slug:
            parts = slug.split("/")
            project = parts[0]
            run = parts[1]
        else:
            project = project or self.settings().get("project")
            if project is None:
                raise CommError("No default project configured.")
            run = run or slug or os.environ.get("WANDB_RUN")
            if run is None:
                run = "latest"
        return (project, run)

    @normalize_exceptions
    def viewer(self):
        query = gql('''
        query Viewer{
            viewer {
                id
                entity
            }
        }
        ''')
        return self.client.execute(query).get('viewer') or {}

    @normalize_exceptions
    def list_projects(self, entity=None):
        """Lists projects in W&B scoped by entity.

        Args:
            entity (str, optional): The entity to scope this project to.  Defaults to public models

        Returns:
                [{"id","name","description"}]
        """
        query = gql('''
        query Models($entity: String!) {
            models(first: 10, entityName: $entity) {
                edges {
                    node {
                        id
                        name
                        description
                    }
                }
            }
        }
        ''')
        return self._flatten_edges(self.client.execute(query, variable_values={
            'entity': entity or self.settings('entity')})['models'])

    @normalize_exceptions
    def list_runs(self, project, entity=None):
        """Lists runs in W&B scoped by project.

        Args:
            project (str): The project to scope the runs to
            entity (str, optional): The entity to scope this project to.  Defaults to public models

        Returns:
                [{"id",name","description"}]
        """
        query = gql('''
        query Buckets($model: String!, $entity: String!) {
            model(name: $model, entityName: $entity) {
                buckets(first: 10) {
                    edges {
                        node {
                            id
                            name
                            description
                        }
                    }
                }
            }
        }
        ''')
        return self._flatten_edges(self.client.execute(query, variable_values={
            'entity': entity or self.settings('entity'),
            'model': project or self.settings('project')})['model']['buckets'])

    @normalize_exceptions
    def run_config(self, project, run=None, entity=None):
        """Get the config for a run

        Args:
            project (str): The project to download, (can include bucket)
            run (str, optional): The run to download
            entity (str, optional): The entity to scope this project to.
        """
        query = gql('''
        query Model($name: String!, $entity: String!, $run: String!) {
            model(name: $name, entityName: $entity) {
                bucket(name: $run) {
                    config
                    commit
                    patch
                }
            }
        }
        ''')

        response = self.client.execute(query, variable_values={
            'name': project, 'bucket': run, 'entity': entity
        })
        run = response['model']['bucket']
        commit = run['commit']
        patch = run['patch']
        config = json.loads(run['config'] or '{}')
        return (commit, config, patch)

    @normalize_exceptions
    def upsert_project(self, project, id=None, description=None, entity=None):
        """Create a new project

        Args:
            project (str): The project to create
            description (str, optional): A description of this project
            entity (str, optional): The entity to scope this project to.
        """
        mutation = gql('''
        mutation UpsertModel($name: String!, $id: String, $entity: String!, $description: String, $repo: String)  {
            upsertModel(input: { id: $id, name: $name, entityName: $entity, description: $description, repo: $repo }) {
                model {
                    name
                    description
                }
            }
        }
        ''')

        response = self.client.execute(mutation, variable_values={
            'name': project, 'entity': entity or self.settings('entity'),
            'description': description, 'repo': self.git.remote_url, 'id': id})
        return response['upsertModel']['model']

    #@normalize_exceptions
    def upsert_run(self, id=None, name=None, project=None, host=None, config=None, description=None, entity=None, commit=None):
        """Update a run

        Args:
            id (str, optional): The existing run to update
            name (str, optional): The name of the run to create
            project (str, optional): The name of the project
            config (dict, optional): The latest config params
            description (str, optional): A description of this project
            entity (str, optional): The entity to scope this project to.
            commit (str, optional): The Git SHA to associate the run with
        """
        mutation = gql('''
        mutation UpsertBucket(
            $id: String, $name: String,
            $project: String,
            $entity: String!,
            $description: String,
            $commit: String,
            $config: JSONString,
            $host: String,
            $debug: Boolean
        ) {
            upsertBucket(input: {
                id: $id, name: $name,
                modelName: $project,
                entityName: $entity,
                description: $description,
                config: $config,
                commit: $commit,
                host: $host,
                debug: $debug
            }) {
                bucket {
                    name
                    description
                    config
                }
            }
        }
        ''')
        if config is not None:
            config = json.dumps(config)
        if not description:
            description = None
        response = self.client.execute(mutation, variable_values={
            'id': id, 'entity': entity or self.settings('entity'), 'name': name, 'project': project,
            'description': description, 'config': config, 'commit': commit or self._commit,
            'host': host, 'debug': os.getenv('DEBUG')})
        return response['upsertBucket']['bucket']

    @normalize_exceptions
    def upload_urls(self, project, files, run=None, entity=None, description=None):
        """Generate temporary resumeable upload urls

        Args:
            project (str): The project to download
            files (list or dict): The filenames to upload
            run (str, optional): The run to upload to
            entity (str, optional): The entity to scope this project to.  Defaults to wandb models

        Returns:
            (bucket_id, file_info)
            bucket_id: id of bucket we uploaded to
            file_info: A dict of filenames and urls, also indicates if this revision already has uploaded files.
                {
                    'weights.h5': { "url": "https://weights.url" },
                    'model.json': { "url": "https://model.json", "updatedAt": '2013-04-26T22:22:23.832Z', 'md5': 'mZFLkyvTelC5g8XnyQrpOw==' },
                }
        """
        query = gql('''
        query Model($name: String!, $files: [String]!, $entity: String!, $run: String!, $description: String) {
            model(name: $name, entityName: $entity) {
                bucket(name: $run, desc: $description) {
                    id
                    files(names: $files) {
                        edges {
                            node {
                                name
                                url(upload: true)
                                updatedAt
                            }
                        }
                    }
                }
            }
        }
        ''')
        query_result = self.client.execute(query, variable_values={
            'name': project, 'run': run or self.settings('run'),
            'entity': entity or self.settings('entity'),
            'description': description,
            'files': [file for file in files]
        })

        run = query_result['model']['bucket']
        result = {file['name']
            : file for file in self._flatten_edges(run['files'])}
        return run['id'], result

    @normalize_exceptions
    def download_urls(self, project, run=None, entity=None):
        """Generate download urls

        Args:
            project (str): The project to download
            run (str, optional): The run to upload to
            entity (str, optional): The entity to scope this project to.  Defaults to wandb models

        Returns:
            A dict of extensions and urls

                {
                    'weights.h5': { "url": "https://weights.url", "updatedAt": '2013-04-26T22:22:23.832Z', 'md5': 'mZFLkyvTelC5g8XnyQrpOw==' },
                    'model.json': { "url": "https://model.url", "updatedAt": '2013-04-26T22:22:23.832Z', 'md5': 'mZFLkyvTelC5g8XnyQrpOw==' }
                }
        """
        query = gql('''
        query Model($name: String!, $entity: String!, $run: String!)  {
            model(name: $name, entityName: $entity) {
                bucket(name: $run) {
                    files {
                        edges {
                            node {
                                name
                                url
                                md5
                                updatedAt
                            }
                        }
                    }
                }
            }
        }
        ''')
        query_result = self.client.execute(query, variable_values={
            'name': project, 'run': run or self.settings('run'),
            'entity': entity or self.settings('entity')})
        files = self._flatten_edges(query_result['model']['bucket']['files'])
        return {file['name']: file for file in files}

    @normalize_exceptions
    def download_file(self, url):
        """Initiate a streaming download

        Args:
            url (str): The url to download

        Returns:
            A tuple of the content length and the streaming response
        """
        response = requests.get(url, stream=True)
        response.raise_for_status()
        return (int(response.headers.get('content-length', 0)), response)

    @normalize_exceptions
    def upload_file(self, url, file, callback=None):
        """Uploads a file to W&B with failure resumption

        Args:
            url (str): The url to download
            file (str): The path to the file you want to upload
            callback (:obj:`func`, optional): A callback which is passed the number of
            bytes uploaded since the last time it was called, used to report progress

        Returns:
            The requests library response object
        """
        attempts = 0
        extra_headers = {}
        if os.stat(file.name).st_size == 0:
            raise CommError("%s is an empty file" % file.name)
        while attempts < self.retries:
            try:
                progress = Progress(file, callback=callback)
                response = requests.put(
                    url, data=progress, headers=extra_headers)
                response.raise_for_status()
                break
            except requests.exceptions.RequestException as e:
                total = progress.len
                status = self._status_request(url, total)
                if status.status_code == 308:
                    attempts += 1
                    completed = int(status.headers['Range'].split("-")[-1])
                    extra_headers = {
                        'Content-Range': 'bytes {completed}-{total}/{total}'.format(
                            completed=completed,
                            total=total
                        ),
                        'Content-Length': str(total - completed)
                    }
                else:
                    raise e
        return response

    @property
    def latest_config(self):
        "The latest config parameters trained on"
        if os.path.exists(__stage_dir__ + 'latest.yaml'):
            config = yaml.load(open(__stage_dir__ + 'latest.yaml'))
            del config['wandb_version']
            return config

    def _md5(self, fname):
        hash_md5 = hashlib.md5()
        with open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return base64.b64encode(hash_md5.digest()).decode('ascii')

    def file_current(self, fname, md5):
        """Checksum a file and compare the md5 with the known md5
        """
        return os.path.isfile(fname) and self._md5(fname) == md5

    @normalize_exceptions
    def pull(self, project, run=None, entity=None):
        """Download files from W&B

        Args:
            project (str): The project to download
            run (str, optional): The run to upload to
            entity (str, optional): The entity to scope this project to.  Defaults to wandb models

        Returns:
            The requests library response object
        """
        project, run = self.parse_slug(project, run=run)
        urls = self.download_urls(project, run, entity)
        responses = []
        for fileName in urls:
            if self.file_current(fileName, urls[fileName]['md5']):
                continue
            with open(fileName, "wb") as file:
                size, res = self.download_file(urls[fileName]['url'])
                responses.append(res)
                for data in res.iter_content():
                    file.write(data)
        return responses

    @normalize_exceptions
    def push(self, project, files, run=None, entity=None, description=None, force=True, progress=False):
        """Uploads multiple files to W&B

        Args:
            project (str): The project to upload to
            files (list or dict): The filenames to upload
            run (str, optional): The run to upload to
            entity (str, optional): The entity to scope this project to.  Defaults to wandb models
            description (str, optional): The description of the changes
            force (bool, optional): Whether to prevent push if git has uncommitted changes

        Returns:
            The requests library response object
        """
        project, run = self.parse_slug(project, run=run)
        # Only tag if enabled
        if self.settings("git_tag"):
            self.tag_and_push(run, description, force)
        run_id, result = self.upload_urls(
            project, files, run, entity, description)
        responses = []
        for file_name, file_info in result.items():
            try:
                open_file = files[file_name] if isinstance(
                    files, dict) else open(file_name, "rb")
            except IOError:
                print("%s does not exist" % file_name)
                continue
            if progress:
                length = os.fstat(open_file.fileno()).st_size
                with click.progressbar(file=progress, length=length, label='Uploading file: %s' % (file_name),
                                       fill_char=click.style('&', fg='green')) as bar:
                    self.upload_file(
                        file_info['url'], open_file, lambda bites: bar.update(bites))
            else:
                responses.append(self.upload_file(file_info['url'], open_file))
            open_file.close()
        if self.latest_config:
            self.upsert_run(id=run_id, description=description,
                            entity=entity, config=self.latest_config)
        return responses

    def get_file_stream_api(self):
        if not self._file_stream_api:
            settings = self.settings()
            if self._current_run_id is None:
                raise UsageError(
                    'Must have a current run to use file stream API.')
            self._file_stream_api = FileStreamApi(
                self.api_key, self.user_agent, settings['base_url'],
                settings['entity'], settings['project'], self._current_run_id)
        return self._file_stream_api

    def tag_and_push(self, name, description, force=True):
        if self.git.enabled and not self.tagged:
            self.tagged = True
            # TODO: this is getting called twice...
            print("Tagging your git repo...")
            if not force and self.git.dirty:
                raise CommError(
                    "You have un-committed changes. Use the force flag or commit your changes.")
            elif self.git.dirty and os.path.exists(__stage_dir__):
                self.git.repo.git.execute(['git', 'diff'], output_stream=open(
                    os.path.join(__stage_dir__, 'diff.patch'), 'wb'))
            self.git.tag(name, description)
            result = self.git.push(name)
            if(result is None or len(result) is None):
                print("Unable to push git tag.")

    def _status_request(self, url, length):
        """Ask google how much we've uploaded"""
        return requests.put(
            url=url,
            headers={'Content-Length': '0',
                     'Content-Range': 'bytes */%i' % length}
        )

    def _flatten_edges(self, response):
        """Return an array from the nested graphql relay structure"""
        return [node['node'] for node in response['edges']]

Chunk = collections.namedtuple('Chunk', ('filename', 'data'))

class DefaultFilePolicy(object):
    def __init__(self):
        self._chunk_id = 0

    def process_chunks(self, chunks):
        chunk_id = self._chunk_id
        self._chunk_id += len(chunks)
        return {
            'offset': chunk_id,
            'content': [c.data for c in chunks]
        }

class CRDedupeFilePolicy(object):
    def __init__(self):
        self._chunk_id = 0

    def process_chunks(self, chunks):
        content = []
        for line in [c.data for c in chunks]:
            if content and content[-1].endswith('\r'):
                content[-1] = line
            else:
                content.append(line)
        chunk_id = self._chunk_id
        self._chunk_id += len(content)
        if content and content[-1].endswith('\r'):
            self._chunk_id -= 1
        return {
            'offset': chunk_id,
            'content': content
        }

class BinaryFilePolicy(object):
    def __init__(self):
        self._offset = 0

    def process_chunks(self, chunks):
        data = b''.join([c.data for c in chunks])
        enc = base64.b64encode(data).decode('ascii')
        offset = self._offset
        self._offset += len(data)
        return {
            'offset': self._offset,
            'content': enc,
            'encoding': 'base64'
        }


class FileStreamApi(object):
    """Pushes chunks of files to our streaming endpoint.

    This class is used as a singleton. It has a thread that serializes access to
    the streaming endpoint and performs rate-limiting and batching.

    TODO: Differentiate between binary/text encoding.
    """
    Finish = collections.namedtuple('Finish', ('failed'))

    HTTP_TIMEOUT = 10
    RATE_LIMIT_SECONDS = 1
    HEARTBEAT_INTERVAL_SECONDS = 15
    MAX_ITEMS_PER_PUSH = 10000

    def __init__(self, api_key, user_agent, base_url, entity, project, run_id):
        self._endpoint = "{base}/{entity}/{project}/{run}/file_stream".format(
            base=base_url,
            entity=entity,
            project=project,
            run=run_id)
        self._client = requests.Session()
        self._client.auth = ('api', api_key)
        self._client.timeout = self.HTTP_TIMEOUT
        self._client.headers.update({
            'User-Agent': user_agent,
        })
        self._file_policies = {}
        self._queue = queue.Queue()
        self._thread = threading.Thread(target=self._thread_body)
        # It seems we need to make this a daemon thread to get sync.py's atexit handler to run, which
        # cleans this thread up.
        self._thread.daemon = True
        self._thread.start()

    def set_file_policy(self, filename, file_policy):
        self._file_policies[filename] = file_policy

    def _read_queue(self):
        # called from the push thread (_thread_body), this does an initial read
        # that'll block for up to RATE_LIMIT_SECONDS. Then it tries to read
        # as much out of the queue as it can. We do this because the http post
        # to the server happens within _thread_body, and can take longer than
        # our rate limit. So next time we get a chance to read the queue we want
        # read all the stuff that queue'd up since last time.
        #
        # If we have more than MAX_ITEMS_PER_PUSH in the queue then the push thread
        # will get behind and data will buffer up in the queue.

        try:
            item = self._queue.get(True, self.RATE_LIMIT_SECONDS)
        except queue.Empty:
            return []
        items = [item]
        for i in range(self.MAX_ITEMS_PER_PUSH):
            try:
                item = self._queue.get_nowait() 
            except queue.Empty:
                return items
            items.append(item)
        return items

    def _thread_body(self):
        posted_data_time = time.time()
        posted_anything_time = time.time()
        ready_chunks = []
        finished = None
        while finished is None:
            items = self._read_queue()
            for item in items:
                if isinstance(item, self.Finish):
                    finished = item
                else:
                    # item is Chunk
                    ready_chunks.append(item)
                
            cur_time = time.time()

            if ready_chunks and cur_time - posted_data_time > self.RATE_LIMIT_SECONDS:
                posted_data_time = cur_time
                posted_anything_time = cur_time
                self._send(ready_chunks)
                ready_chunks = []

            if cur_time - posted_anything_time > self.HEARTBEAT_INTERVAL_SECONDS:
                logger.debug("Sending heartbeat at %s", cur_time)
                posted_anything_time = cur_time
                util.request_with_retry(self._client.post,
                                        self._endpoint, json={'complete': False, 'failed': False})

        # post the final close message. (item is self.Finish instance now)
        util.request_with_retry(self._client.post,
                                self._endpoint, json={'complete': True, 'failed': bool(finished.failed)})

    def _send(self, chunks):
        # create files dict. dict of <filename: chunks> pairs where chunks is a list of
        # [chunk_id, chunk_data] tuples (as lists since this will be json).
        files = {}
        for filename, file_chunks in itertools.groupby(chunks, lambda c: c.filename):
            file_chunks = list(file_chunks)  # groupby returns iterator
            if filename not in self._file_policies:
                self._file_policies[filename] = DefaultFilePolicy()
            files[filename] = self._file_policies[filename].process_chunks(file_chunks)

        util.request_with_retry(
            self._client.post, self._endpoint, json={'files': files})

    def push(self, filename, data):
        """Push a chunk of a file to the streaming endpoint.

        Args:
            filename: Name of file that this is a chunk of.
            chunk_id: TODO: change to 'offset'
            chunk: File data.
        """
        self._queue.put(Chunk(filename, data))

    def finish(self, failed):
        """Cleans up.

        Anything pushed after finish will be dropped.

        Args:
            failed: Set True to to display run failure in UI.
        """
        self._queue.put(self.Finish(failed))
        self._thread.join()



from six import BytesIO
from wandb import Api
import base64
import csv
import tempfile
import os

def results(project=None, run=None):
    r = Results(project, run)
    try:
        yield r
    finally:
        r.close()

class Results(object):
    """Generates results to be compared in WandB

    with Results("project/run") as r:
        for truth, img_data in test_data:
            label, score = model.predict(img_data)
            img = array_to_img(img_data)
            r.write(input=r.encode_image(img), output=label,
                truth=truth, score=score)
    """
    def __init__(self, project=None, run=None):
        self.api = Api()
        self.project = project or self.api.settings("project")
        self.run = run or os.getenv("WANDB_RUN")
        self.tempfile = tempfile.NamedTemporaryFile(mode='w')
        self.csv = csv.writer(self.tempfile)
        self.csv.writerow(["input","output","probability","truth","loss"])
        self.rows = 0

    def __enter__(self):
        return self

    def __exit__(self, kind, value, extra):
        self.close()

    def encode_image(self, img, format="png"):
        """Accepts a PIL image and returns an encoded data uri"""
        buffer = BytesIO()
        img.save(buffer, format=format)
        return self.encode_data(buffer.getvalue(), format="image/%s" % format)

    def encode_data(self, data, format="image/png"):
        """Creates a data uri from raw data"""
        return "data:{format};base64,{img}".format(
            format=format,
            img=base64.b64encode(data).decode("UTF-8")
        )

    def write(self, **kwargs):
        self.rows += 1
        self.csv.writerow(
            [kwargs["input"], kwargs["output"], kwargs.get("probability"),
                kwargs["truth"], kwargs["loss"]])

    def close(self):
        self.tempfile.flush()
        self.api.push(self.project, {'results.csv': open(self.tempfile.name, "rb")},
            run=self.run)
        self.tempfile.close()







# New tensorflow file writer that also updates wandb
#
# Usage:
# history = wandb.History()
# summary = wandb.Summary()
#
# summary_writer = WandbFileWriter(FLAGS.log_dir, wandb_summary=summary, wandb_history=history)
#
#     slim.learning.train(
#        train_op,
#        FLAGS.log_dir,
#        summary_writer=summary_writer)
#

from tensorflow.python.summary import summary
from tensorflow.python.summary.writer.event_file_writer import EventFileWriter
import wandb


class WandbEventFileWriter(EventFileWriter):
  def __init__(self, logdir, max_queue=10, flush_secs=120,
               filename_suffix=None, wandb_summary=None, wandb_history=None):
    self._wandb_summary = wandb_summary
    self._wandb_history = wandb_history
    super(WandbEventFileWriter, self).__init__(logdir, max_queue, flush_secs,
               filename_suffix)


  def add_event(self, event):
    for v in event.summary.value:
      if v.simple_value:
        print("Tag: " + v.tag + " Val: " + str(v.simple_value))
        self._wandb_summary[v.tag] = v.simple_value

    super(WandbEventFileWriter, self).add_event(event)


class WandbFileWriter(summary.FileWriter):
  def __init__(self,
               logdir,
               graph=None,
               max_queue=10,
               flush_secs=120,
               graph_def=None,
               filename_suffix=None,
               wandb_summary=None,
               wandb_history=None):

    event_writer = WandbEventFileWriter(logdir, max_queue, flush_secs,
                                    filename_suffix, wandb_summary, wandb_history)

    super(summary.FileWriter, self).__init__(event_writer, graph, graph_def)



import csv
import os

from wandb import util

HISTORY_FNAME = 'wandb-history.jsonl'

class History(object):
    """Used to store data that changes over time during runs. """
    def __init__(self, out_dir='.'):
        self.fname = os.path.join(out_dir, HISTORY_FNAME)
        self._file = open(self.fname, 'w')
        self.rows = []

    def add(self, row):
        self.rows.append(row)
        self._file.write(util.json_dumps_safer(row))
        self._file.write('\n')
        self._file.flush()


# -*- coding: utf-8 -*-

__author__ = """Chris Van Pelt"""
__email__ = 'vanpelt@wandb.com'
__version__ = '0.4.20'

import types
import sys
import logging
import os

# We use the hidden version if it already exists, otherwise non-hidden.
if os.path.exists('.wandb'):
    __stage_dir__ = '.wandb/'
elif os.path.exists('wandb'):
    __stage_dir__ = "wandb/"
else:
    __stage_dir__ = None

from .git_repo import GitRepo
from .api import Api, Error
from .sync import Sync
from .config import Config
from .results import Results
from .summary import Summary
from .history import History
from .keras import WandBKerasCallback
from wandb import wandb_run

# Three possible modes:
#     'cli': running from "wandb" command
#     'run': we're a script launched by "wandb run"
#     'dryrun': we're a script not launched by "wandb run"
MODE = os.environ.get('WANDB_MODE', 'dryrun')

# The current run (a Run object)
run = None

if MODE == 'run':
    run = wandb_run.Run(os.getenv('WANDB_RUN_ID'),
                        os.getenv('WANDB_RUN_DIR'),
                        Config())
elif MODE == 'dryrun':
    run_id = wandb_run.generate_id()
    run = wandb_run.Run(run_id, wandb_run.run_dir_path(
        run_id, dry=True), Config())


# called by cli.py
# Even when running the wandb cli, __init__.py is imported before main() runs, so we set
# cli mode afterward. This means there's a period of time before this call when MODE will
# be dryrun
def _set_cli_mode():
    global MODE, run
    MODE = 'cli'
    run = None


if __stage_dir__ is not None:
    log_fname = __stage_dir__ + 'debug.log'
else:
    log_fname = './wandb-debug.log'
logging.basicConfig(
    filemode="w",
    filename=log_fname,
    level=logging.DEBUG)


def push(*args, **kwargs):
    Api().push(*args, **kwargs)


def pull(*args, **kwargs):
    Api().pull(*args, **kwargs)


def sync(extra_config=None):
    if MODE == 'run':
        api = Api()
        if api.api_key is None:
            raise Error(
                "No API key found, run `wandb login` or set WANDB_API_KEY")
        api.set_current_run_id(run.id)
        if extra_config is not None:
            run.config.update(extra_config)
        sync = Sync(api, run.id, config=run.config)
        sync.watch(files='*')
    elif MODE == 'dryrun':
        print('wandb dryrun mode. Use "wandb run <script>" to save results to wandb.\n'
              'Run directory: %s' % run.dir)
    elif MODE == 'cli':
        raise Error('wandb.sync called from cli mode')


__all__ = ["Api", "Error", "Config", "Results", "History", "Summary",
           "WandBKerasCallback"]
